package GEmeTool::Options;

use v5.16.3;

use strict;
use warnings;

use Moo;
use GEmeTool::DB;

my $db = GEmeTool::DB->connect;

my $query_get_option = <<'EOF';
SELECT value
FROM options
WHERE name = ?;
EOF

my $query_replace_option = <<'EOF';
INSERT OR REPLACE 
INTO options (name, value)
VALUES (?, ?);
EOF

sub _get_option_value {
    my $self = shift;
    my $name = shift;
    my $result = $db->selectrow_hashref($query_get_option, {}, $name);
    if (!defined $result) {
        return;
    }
    return $result->{value};
}

sub _set_option_value {
    my $self = shift;
    my $name = shift;
    my $value = shift;
    return $db->do($query_replace_option, {}, $name, $value);
}

sub get_last_dir_open {
    my $self = shift;
    return $self->_get_option_value('last_dir_open');
}

sub set_last_dir_open {
    my $self = shift;
    my $value = shift;
    die '$value must be defined' if !defined $value;
    return $self->_set_option_value('last_dir_open', $value);
}

sub get_last_dir_save {
    my $self = shift;
    return $self->_get_option_value('last_dir_save') // $self->get_last_dir_open;
}

sub set_last_dir_save {
    my $self = shift;
    my $value = shift;
    die '$value must be defined' if !defined $value;
    return $self->_set_option_value('last_dir_save', $value);
}
1;
